@@ -1,5 +1,7 @@
 from collections import defaultdict, deque, namedtuple
 import collections.abc
+from dataclasses import dataclass
+from typing import Optional
 from distutils.version import LooseVersion
 import copy
 import json
@@ -9,6 +11,7 @@ import itertools
 import inspect
 import os
 from pkg_resources import resource_filename as rs_fn
+from importlib.metadata import version
 import threading
 import time as ttime
 import types
@@ -33,6 +36,8 @@ class DocumentNames(Enum):
     resource = 'resource'
     event_page = 'event_page'
     datum_page = 'datum_page'
+    stream_resource = 'stream_resource'
+    stream_datum = 'stream_datum'
     bulk_datum = 'bulk_datum'  # deprecated
     bulk_events = 'bulk_events'  # deprecated
 
@@ -203,6 +208,12 @@ class DocumentRouter:
     def datum_page(self, doc):
         return NotImplemented
 
+    def stream_datum(self, doc):
+        return NotImplemented
+
+    def stream_resource(self, doc):
+        return NotImplemented
+
     def bulk_events(self, doc):
         # Do not modify this in a subclass. Use event_page.
         warnings.warn(
@@ -486,6 +497,10 @@ class Filler(DocumentRouter):
         A cache of Datum documents. If None, a dict is used.
     descriptor_cache : dict, optional
         A cache of EventDescriptor documents. If None, a dict is used.
+    stream_resource_cache : dict, optional
+        A cache of StreamResource documents. If None, a dict is used.
+    stream_datum_cache : dict, optional
+        A cache of StreamDatum documents. If None, a dict is used.
     retry_intervals : Iterable, optional
         If data is not found on the first try, there may a race between the
         I/O systems creating the external data and this stream of Documents
@@ -526,7 +541,8 @@ class Filler(DocumentRouter):
     def __init__(self, handler_registry, *,
                  include=None, exclude=None, root_map=None, coerce='as_is',
                  handler_cache=None, resource_cache=None, datum_cache=None,
-                 descriptor_cache=None, inplace=None,
+                 descriptor_cache=None, stream_resource_cache=None,
+                 stream_datum_cache=None, inplace=None,
                  retry_intervals=(0.001, 0.002, 0.004, 0.008, 0.016, 0.032,
                                   0.064, 0.128, 0.256, 0.512, 1.024)):
         if inplace is None:
@@ -579,10 +595,16 @@ class Filler(DocumentRouter):
             datum_cache = self.get_default_datum_cache()
         if descriptor_cache is None:
             descriptor_cache = self.get_default_descriptor_cache()
+        if stream_resource_cache is None:
+            stream_resource_cache = self.get_default_stream_resource_cache()
+        if stream_datum_cache is None:
+            stream_datum_cache = self.get_default_stream_datum_cache()
         self._handler_cache = handler_cache
         self._resource_cache = resource_cache
         self._datum_cache = datum_cache
         self._descriptor_cache = descriptor_cache
+        self._stream_resource_cache = stream_resource_cache
+        self._stream_datum_cache = stream_datum_cache
         if retry_intervals is None:
             retry_intervals = []
         self.retry_intervals = retry_intervals
@@ -600,6 +622,8 @@ class Filler(DocumentRouter):
             type(self._resource_cache) is type(other._resource_cache) and
             type(self._datum_cache) is type(other._datum_cache) and
             type(self._descriptor_cache) is type(other._descriptor_cache) and
+            type(self._stream_resource_cache) is type(other._stream_resource_cache) and
+            type(self._stream_datum_cache) is type(other._stream_datum_cache) and
             self.retry_intervals == other.retry_intervals
         )
 
@@ -615,6 +639,8 @@ class Filler(DocumentRouter):
             resource_cache=self._resource_cache,
             datum_cache=self._datum_cache,
             descriptor_cache=self._descriptor_cache,
+            stream_resource_cache=self._stream_resource_cache,
+            stream_datum_cache=self._stream_datum_cache,
             retry_intervals=self.retry_intervals)
 
     def __setstate__(self, d):
@@ -637,6 +663,8 @@ class Filler(DocumentRouter):
         self._resource_cache = d['resource_cache']
         self._datum_cache = d['datum_cache']
         self._descriptor_cache = d['descriptor_cache']
+        self._stream_resource_cache = d['stream_resource_cache']
+        self._stream_datum_cache = d['stream_datum_cache']
         retry_intervals = d['retry_intervals']
         if retry_intervals is None:
             retry_intervals = []
@@ -670,6 +698,14 @@ class Filler(DocumentRouter):
     def get_default_handler_cache():
         return {}
 
+    @staticmethod
+    def get_default_stream_datum_cache():
+        return {}
+
+    @staticmethod
+    def get_default_stream_resource_cache():
+        return {}
+
     @property
     def inplace(self):
         return self._inplace
@@ -677,7 +713,8 @@ class Filler(DocumentRouter):
     def clone(self, handler_registry=None, *,
               root_map=None, coerce=None,
               handler_cache=None, resource_cache=None, datum_cache=None,
-              descriptor_cache=None, inplace=None,
+              descriptor_cache=None, stream_resource_cache=None,
+              stream_datum_cache=None, inplace=None,
               retry_intervals=None):
         """
         Create a new Filler instance from this one.
@@ -704,6 +741,8 @@ class Filler(DocumentRouter):
                       resource_cache=resource_cache,
                       datum_cache=datum_cache,
                       descriptor_cache=descriptor_cache,
+                      stream_resource_cache=stream_resource_cache,
+                      stream_datum_cache=stream_datum_cache,
                       inplace=inplace,
                       retry_intervals=retry_intervals)
 
@@ -788,6 +827,13 @@ class Filler(DocumentRouter):
         self._datum_cache[doc['datum_id']] = doc
         return doc
 
+    def stream_resource(self, doc):
+        self._stream_resource_cache[doc["uid"]] = doc
+        return doc
+
+    def stream_datum(self, doc):
+        self._stream_datum_cache[doc["uid"]] = doc
+
     def event_page(self, doc):
         # TODO We may be able to fill a page in place, and that may be more
         # efficient than unpacking the page in to Events, filling them, and the
@@ -1268,6 +1314,7 @@ class RunRouter(DocumentRouter):
 
         # Map Resource UID to RunStart UID.
         self._resources = {}
+        self._stream_resources = {}
 
         # Old-style Resources that do not have a RunStart UID
         self._unlabeled_resources = deque(maxlen=10000)
@@ -1317,19 +1364,27 @@ class RunRouter(DocumentRouter):
             self._subfactories[uid].extend(subfactories)
 
     def descriptor(self, descriptor_doc):
-        uid = descriptor_doc['uid']
+        descriptor_uid = descriptor_doc['uid']
         start_uid = descriptor_doc['run_start']
+
+        # Keep track of the RunStart UID -> [EventDescriptor UIDs] mapping for
+        # purposes of cleanup in stop().
+        self._start_to_descriptors[start_uid].append(descriptor_uid)
+        # Keep track of the EventDescriptor UID -> RunStartUID for filling
+        # purposes.
+        self._descriptor_to_start[descriptor_uid] = start_uid
+
         self._fillers[start_uid].descriptor(descriptor_doc)
         # Apply all factory cbs for this run to this descriptor, and run them.
         factory_cbs = self._factory_cbs_by_start[start_uid]
-        self._factory_cbs_by_descriptor[uid].extend(factory_cbs)
+        self._factory_cbs_by_descriptor[descriptor_uid].extend(factory_cbs)
         for callback in factory_cbs:
             callback('descriptor', descriptor_doc)
         # Let all the subfactories add any relevant callbacks.
         for subfactory in self._subfactories[start_uid]:
             callbacks = subfactory('descriptor', descriptor_doc)
             self._subfactory_cbs_by_start[start_uid].extend(callbacks)
-            self._subfactory_cbs_by_descriptor[uid].extend(callbacks)
+            self._subfactory_cbs_by_descriptor[descriptor_uid].extend(callbacks)
             for callback in callbacks:
                 try:
                     start_doc = self._start_to_start_doc[start_uid]
@@ -1346,12 +1401,6 @@ class RunRouter(DocumentRouter):
                         DOCS_PASSED_IN_1_14_0_WARNING.format(
                             callback=callback, name='descriptor', err=err))
                     raise err
-        # Keep track of the RunStart UID -> [EventDescriptor UIDs] mapping for
-        # purposes of cleanup in stop().
-        self._start_to_descriptors[start_uid].append(uid)
-        # Keep track of the EventDescriptor UID -> RunStartUID for filling
-        # purposes.
-        self._descriptor_to_start[uid] = start_uid
 
     def event_page(self, doc):
         descriptor_uid = doc['descriptor']
@@ -1396,6 +1445,15 @@ class RunRouter(DocumentRouter):
             for callback in self._subfactory_cbs_by_start[start_uid]:
                 callback('datum_page', doc)
 
+    def stream_datum(self, doc):
+        resource_uid = doc["stream_resource"]
+        start_uid = self._stream_resources[resource_uid]
+        self._fillers[start_uid].stream_datum(doc)
+        for callback in self._factory_cbs_by_start[start_uid]:
+            callback("stream_datum", doc)
+        for callback in self._subfactory_cbs_by_start[start_uid]:
+            callback("stream_datum", doc)
+
     def resource(self, doc):
         try:
             start_uid = doc['run_start']
@@ -1421,6 +1479,15 @@ class RunRouter(DocumentRouter):
             for callback in self._subfactory_cbs_by_start[start_uid]:
                 callback('resource', doc)
 
+    def stream_resource(self, doc):
+        start_uid = doc["run_start"]  # No need for Try
+        self._fillers[start_uid].stream_resource(doc)
+        self._stream_resources[doc["uid"]] = doc["run_start"]
+        for callback in self._factory_cbs_by_start[start_uid]:
+            callback("stream_resource", doc)
+        for callback in self._subfactory_cbs_by_start[start_uid]:
+            callback("stream_resource", doc)
+
     def stop(self, doc):
         start_uid = doc['run_start']
         for callback in self._factory_cbs_by_start[start_uid]:
@@ -1521,6 +1588,8 @@ SCHEMA_NAMES = {DocumentNames.start: 'schemas/run_start.json',
                 DocumentNames.datum: 'schemas/datum.json',
                 DocumentNames.datum_page: 'schemas/datum_page.json',
                 DocumentNames.resource: 'schemas/resource.json',
+                DocumentNames.stream_datum: 'schemas/stream_datum.json',
+                DocumentNames.stream_resource: 'schemas/stream_resource.json',
                 # DEPRECATED:
                 DocumentNames.bulk_events: 'schemas/bulk_events.json',
                 DocumentNames.bulk_datum: 'schemas/bulk_datum.json'}
@@ -1529,11 +1598,10 @@ for name, filename in SCHEMA_NAMES.items():
     with open(rs_fn('event_model', filename)) as fin:
         schemas[name] = json.load(fin)
 
-
 # We pin jsonschema >=3.0.0 in requirements.txt but due to pip's dependency
 # resolution it is easy to end up with an environment where that pin is not
 # respected. Thus, we maintain best-effort support for 2.x.
-if LooseVersion(jsonschema.__version__) >= LooseVersion("3.0.0"):
+if LooseVersion(version("jsonschema")) >= LooseVersion("3.0.0"):
     def _is_array(checker, instance):
         return (
             jsonschema.validators.Draft7Validator.TYPE_CHECKER.is_type(instance, 'array') or
@@ -1561,13 +1629,29 @@ else:
 __version__ = get_versions()['version']
 del get_versions
 
-ComposeRunBundle = namedtuple('ComposeRunBundle',
-                              'start_doc compose_descriptor compose_resource '
-                              'compose_stop')
+
+@dataclass
+class ComposeRunBundle:
+    """Extensible compose run bundle. This maintains backward compatibility by unpacking into a basic
+    run bundle (start, compose_descriptor, compose_resource, stop). Further extensions are optional and
+    require keyword referencing (i.e. compose_stream_resource).
+    """
+    start_doc: dict
+    compose_descriptor: callable
+    compose_resource: callable
+    compose_stop: callable
+    compose_stream_resource: Optional[callable] = None
+
+    def __iter__(self):
+        return iter((self.start_doc, self.compose_descriptor, self.compose_resource, self.compose_stop))
+
+
 ComposeDescriptorBundle = namedtuple('ComposeDescriptorBundle',
                                      'descriptor_doc compose_event compose_event_page')
 ComposeResourceBundle = namedtuple('ComposeResourceBundle',
                                    'resource_doc compose_datum compose_datum_page')
+ComposeStreamResourceBundle = namedtuple('ComposeStreamResourceBundle',
+                                         'stream_resource_doc compose_stream_data')
 
 
 def compose_datum(*, resource, counter, datum_kwargs, validate=True):
@@ -1595,18 +1679,20 @@ def compose_datum_page(*, resource, counter, datum_kwargs, validate=True):
 default_path_semantics = {'posix': 'posix', 'nt': 'windows'}[os.name]
 
 
-def compose_resource(*, start, spec, root, resource_path, resource_kwargs,
-                     path_semantics=default_path_semantics, uid=None, validate=True):
+def compose_resource(*, spec, root, resource_path, resource_kwargs,
+                     path_semantics=default_path_semantics, start=None, uid=None, validate=True):
     if uid is None:
         uid = str(uuid.uuid4())
     counter = itertools.count()
     doc = {'uid': uid,
-           'run_start': start['uid'],
            'spec': spec,
            'root': root,
            'resource_path': resource_path,
            'resource_kwargs': resource_kwargs,
            'path_semantics': path_semantics}
+    if start:
+        doc['run_start'] = start['uid']
+
     if validate:
         schema_validators[DocumentNames.resource].validate(doc)
 
@@ -1616,6 +1702,56 @@ def compose_resource(*, start, spec, root, resource_path, resource_kwargs,
         partial(compose_datum_page, resource=doc, counter=counter))
 
 
+def compose_stream_datum(*, stream_resource, stream_name, counter, datum_kwargs,
+                         event_count=1, event_offset=0, validate=True):
+    resource_uid = stream_resource['uid']
+    if stream_name not in stream_resource["stream_names"]:
+        raise EventModelKeyError("Attempt to create stream_datum with name not included in stream_resource")
+    block_idx = next(counter)
+    doc = dict(stream_resource=resource_uid,
+               datum_kwargs=datum_kwargs,
+               uid=f"{resource_uid}/{stream_name}/{block_idx}",
+               stream_name=stream_name,
+               block_idx=block_idx,
+               event_count=event_count,
+               event_offset=event_offset,
+               )
+    if validate:
+        schema_validators[DocumentNames.stream_datum].validate(doc)
+    return doc
+
+
+def compose_stream_resource(*, spec, root, resource_path, resource_kwargs, stream_names, counters=(),
+                            path_semantics=default_path_semantics, start=None, uid=None, validate=True):
+    if uid is None:
+        uid = str(uuid.uuid4())
+    if isinstance(stream_names, str):
+        stream_names = [stream_names, ]
+    if len(counters) == 0:
+        counters = [itertools.count() for _ in stream_names]
+    elif len(counters) > len(stream_names):
+        raise ValueError(f"Insufficient number of counters {len(counters)} for stream names: {stream_names}")
+
+    doc = dict(uid=uid,
+               spec=spec,
+               root=root,
+               resource_path=resource_path,
+               resource_kwargs=resource_kwargs,
+               stream_names=stream_names,
+               path_semantics=path_semantics)
+    if start:
+        doc["run_start"] = start["uid"]
+
+    if validate:
+        schema_validators[DocumentNames.stream_resource].validate(doc)
+
+    return ComposeStreamResourceBundle(
+        doc,
+        [partial(compose_stream_datum, stream_resource=doc, stream_name=stream_name, counter=counter) for
+         stream_name, counter in zip(stream_names, counters)]
+    )
+
+
 def compose_stop(*, start, event_counter, poison_pill,
                  exit_status='success', reason='',
                  uid=None, time=None,
@@ -1787,7 +1923,9 @@ def compose_run(*, uid=None, time=None, metadata=None, validate=True):
                 event_counter=event_counter),
         partial(compose_resource, start=doc),
         partial(compose_stop, start=doc, event_counter=event_counter,
-                poison_pill=poison_pill))
+                poison_pill=poison_pill),
+        compose_stream_resource=partial(compose_stream_resource, start=doc)
+    )
 
 
 def pack_event_page(*events):
